#!/usr/bin/env python
"""Determine package version from git repository tag

Each time this file is imported it checks whether the package version
can be determined using `git describe`. If this fails (because either
this file is not located at the 1st level down the repository root or
it is not under version control), the version is read from the script
"_version_save.py" which is not versioned by git, but always included
in the final distribution archive (e.g. via PyPI). If the git version
does not match the saved version, then "_version_save.py" is updated.


Usage
-----
1. Put this file in your main module directory:

    REPO_ROOT/package_name/_version.py

2. Add this line to REPO_ROOT/package_name/__init__.py

    from ._version import version as __version__  # noqa: F401

3. (Optional) Add this line to REPO_ROOT/.gitignore

    _version_save.py

Features
--------
- supports Python 2 and 3
- supports frozen applications (e.g. PyInstaller)
- supports installing into a virtual environment that is located in
  a git repository
- saved version is located in a python file and therefore no other
  files (e.g. MANIFEST.in) need be edited
- fallback version is the creation date
- excluded from code coverage via "pragma: no cover"

Changelog
---------
2019-11-06.2
 - use os.path.split instead of counting os.path.sep (Windows)
2019-11-06
 - remove deprecated imp dependency (replace with parser)
 - check whether this file is versioned and its location is correct
 - code cleanup and docs update
"""
from __future__ import print_function

# Put the entire script into a `True` statement and add the hint
# `pragma: no cover` to ignore code coverage here.
if True:  # pragma: no cover
    import os
    from os.path import abspath, basename, dirname, join, split
    import subprocess
    import sys
    import time
    import traceback
    import warnings

    def git_describe():
        """
        Return a string describing the version returned by the
        command `git describe --tags HEAD`.
        If it is not possible to determine the correct version,
        then an empty string is returned.
        """
        # Make sure we are in a directory that belongs to the correct
        # repository.
        ourdir = dirname(abspath(__file__))

        def _minimal_ext_cmd(cmd):
            # Construct minimal environment
            env = {}
            for k in ['SYSTEMROOT', 'PATH']:
                v = os.environ.get(k)
                if v is not None:
                    env[k] = v
            # LANGUAGE is used on win32
            env['LANGUAGE'] = 'C'
            env['LANG'] = 'C'
            env['LC_ALL'] = 'C'
            pop = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   env=env)
            out = pop.communicate()[0]
            return out.strip().decode('ascii', errors="ignore")

        # change directory
        olddir = abspath(os.curdir)
        os.chdir(ourdir)

        # Make sure that we are getting "git describe" from our own
        # repository (and not from a repository where we just happen
        # to be in the directory tree).
        git_revision = ""
        try:
            # If this file is not under version control, "loc" will
            # be empty.
            loc = _minimal_ext_cmd(['git', 'ls-files', '--full-name',
                                    __file__])
            # If it is under version control, it should be located
            # one hierarchy down from the repository root (either
            # __file__ is "docs/conf.py" or "package_name/_version.py".
            if len(split(loc)) == 2:
                try:
                    git_revision = _minimal_ext_cmd(['git', 'describe',
                                                     '--tags', 'HEAD'])
                except OSError:
                    pass
        except OSError:
            pass
        # Go back to original directory
        os.chdir(olddir)

        return git_revision

    def load_version(versionfile):
        """load version from version_save.py"""
        longversion = ""
        try:
            with open(versionfile, "r") as fd:
                data = fd.readlines()
                for line in data:
                    if line.startswith("longversion"):
                        longversion = line.split("=")[1].strip().strip("'")
        except BaseException:
            try:
                from ._version_save import longversion
            except BaseException:
                try:
                    from _version_save import longversion
                except BaseException:
                    pass

        return longversion

    def write_version(version, versionfile):
        """save version to version_save.py"""
        data = "#!/usr/bin/env python\n" \
            + "# This file was created automatically\n" \
            + "longversion = '{VERSION}'\n"
        try:
            with open(versionfile, "w") as fd:
                fd.write(data.format(VERSION=version))
        except BaseException:
            if not os.path.exists(versionfile):
                # Only issue a warning if the file does not exist.
                msg = "Could not write package version to {}.".format(
                    versionfile)
                warnings.warn(msg)

    hdir = dirname(abspath(__file__))
    if basename(__file__) == "conf.py" and "name" in locals():
        # This script is executed in conf.py from the docs directory
        versionfile = join(join(join(hdir, ".."),
                                name),  # noqa: F821
                           "_version_save.py")
    else:
        # This script is imported as a module
        versionfile = join(hdir, "_version_save.py")

    # Determine the accurate version
    longversion = ""

    # 1. git describe
    try:
        # Get the version using `git describe`
        longversion = git_describe()
    except BaseException:
        pass

    # 2. previously created version file
    if longversion == "":
        # Either this is not a git repository or we are in the
        # wrong git repository.
        # Get the version from the previously generated `_version_save.py`
        longversion = load_version(versionfile)

    # 3. last resort: date
    if longversion == "":
        print("Could not determine version. Reason:")
        print(traceback.format_exc())
        ctime = os.stat(__file__)[8]
        longversion = time.strftime("%Y.%m.%d-%H-%M-%S", time.gmtime(ctime))
        print("Using creation time as version: {}".format(longversion))

    if not hasattr(sys, 'frozen'):
        # Save the version to `_version_save.py` to allow distribution using
        # `python setup.py sdist`.
        # This is only done if the program is not frozen (with e.g.
        # pyinstaller),
        if longversion != load_version(versionfile):
            write_version(longversion, versionfile)

    # PEP 440-conform development version:
    version = ".post".join(longversion.split("-")[:2])



import datetime
import time

import ckan.model as model

import click


from . import jobs


def click_echo(message, am_on_a_new_line):
    if not am_on_a_new_line:
        click.echo("")
    click.echo(message)


@click.option('--modified-days', default=-1,
              help='Only run for datasets modified within this number of days '
                   + 'in the past. Set to -1 to apply to all datasets.')
@click.option('--force', help="Regenerate preview for all resources",
              is_flag=True)
@click.command()
def run_jobs_dc_view(modified_days=-1, force=False):
    """Generate preview image for all RT-DC resources

    This also happens for draft datasets.
    """
    # go through all datasets
    datasets = model.Session.query(model.Package)

    if modified_days >= 0:
        # Search only the last `days` days.
        past = datetime.date.today() - datetime.timedelta(days=modified_days)
        past_str = time.strftime("%Y-%m-%d", past.timetuple())
        datasets = datasets.filter(model.Package.metadata_modified >= past_str)

    nl = False  # new line character
    for dataset in datasets:
        nl = False
        click.echo(f"Checking dataset {dataset.id}\r", nl=False)
        for resource in dataset.resources:
            res_dict = resource.as_dict()
            try:
                if jobs.create_preview_job(res_dict, override=force):
                    click_echo(f"Created preview for {resource.name}", nl)
                    nl = True
            except KeyboardInterrupt:
                raise
            except BaseException as e:
                click_echo(
                    f"{e.__class__.__name__}: {e} for {res_dict['name']}", nl)
                nl = True
    if not nl:
        click.echo("")
    click.echo("Done!")


def get_commands():
    return [run_jobs_dc_view]



"""Testing background jobs

Due to the asynchronous nature of background jobs, code that uses them needs
to be handled specially when writing tests.

A common approach is to use the mock package to replace the
ckan.plugins.toolkit.enqueue_job function with a mock that executes jobs
synchronously instead of asynchronously
"""
import mock
import pathlib
import time

import pytest
import requests

import ckan.lib
import ckan.tests.factories as factories

import ckanext.dcor_schemas.plugin
import dcor_shared

from .helper_methods import make_dataset


data_dir = pathlib.Path(__file__).parent / "data"


def synchronous_enqueue_job(job_func, args=None, kwargs=None, title=None,
                            queue=None, rq_kwargs=None):
    """
    Synchronous mock for ``ckan.plugins.toolkit.enqueue_job``.
    """
    if rq_kwargs is None:
        rq_kwargs = {}
    args = args or []
    kwargs = kwargs or {}
    job_func(*args, **kwargs)


# We need the dcor_depot extension to make sure that the symbolic-
# linking pipeline is used.
@pytest.mark.ckan_config('ckan.plugins', 'dcor_depot dcor_schemas dc_view')
@pytest.mark.usefixtures('clean_db', 'with_request_context')
@mock.patch('ckan.plugins.toolkit.enqueue_job',
            side_effect=synchronous_enqueue_job)
def test_create_preview_job(enqueue_job_mock, create_with_upload, monkeypatch,
                            ckan_config, tmpdir):
    monkeypatch.setitem(ckan_config, 'ckan.storage_path', str(tmpdir))
    monkeypatch.setattr(ckan.lib.uploader,
                        'get_storage_path',
                        lambda: str(tmpdir))
    monkeypatch.setattr(
        ckanext.dcor_schemas.plugin,
        'DISABLE_AFTER_DATASET_CREATE_FOR_CONCURRENT_JOB_TESTS',
        True)

    user = factories.User()
    owner_org = factories.Organization(users=[{
        'name': user['id'],
        'capacity': 'admin'
    }])
    # Note: `call_action` bypasses authorization!
    # create 1st dataset
    create_context = {'ignore_auth': False, 'user': user['name'],
                      'api_version': 3}
    dataset = make_dataset(create_context,
                           owner_org,
                           activate=False)
    path = data_dir / "calibration_beads_47.rtdc"
    content = path.read_bytes()
    result = create_with_upload(
        content, 'test.rtdc',
        url="upload",
        package_id=dataset["id"],
        context=create_context,
    )
    resource_path = dcor_shared.get_resource_path(result["id"])
    assert resource_path.exists()
    preview_path = resource_path.with_name(resource_path.name + "_preview.jpg")
    # give the background job a little time to complete
    for ii in range(100):
        if not preview_path.exists():
            time.sleep(0.1)
        else:
            assert preview_path.stat().st_size > 1000
            break
    else:
        raise ValueError("Preview generation timed out after 10s!")


# We need the dcor_depot extension to make sure that the symbolic-
# linking pipeline is used.
@pytest.mark.ckan_config('ckan.plugins', 'dcor_depot dc_view dcor_schemas')
@pytest.mark.usefixtures('clean_db', 'with_request_context')
@mock.patch('ckan.plugins.toolkit.enqueue_job',
            side_effect=synchronous_enqueue_job)
def test_upload_preview_dataset_to_s3_job(
        enqueue_job_mock, create_with_upload, monkeypatch, ckan_config,
        tmpdir):
    monkeypatch.setitem(ckan_config, 'ckan.storage_path', str(tmpdir))
    monkeypatch.setattr(ckan.lib.uploader,
                        'get_storage_path',
                        lambda: str(tmpdir))
    monkeypatch.setattr(
        ckanext.dcor_schemas.plugin,
        'DISABLE_AFTER_DATASET_CREATE_FOR_CONCURRENT_JOB_TESTS',
        True)

    user = factories.User()
    owner_org = factories.Organization(users=[{
        'name': user['id'],
        'capacity': 'admin'
    }])
    # Note: `call_action` bypasses authorization!
    # create 1st dataset
    create_context = {'ignore_auth': False,
                      'user': user['name'],
                      'api_version': 3}
    ds_dict, res_dict = make_dataset(create_context,
                                     owner_org,
                                     create_with_upload=create_with_upload,
                                     activate=True)
    bucket_name = dcor_shared.get_ckan_config_option(
        "dcor_object_store.bucket_name").format(
        organization_id=ds_dict["organization"]["id"])
    rid = res_dict["id"]
    object_name = f"preview/{rid[:3]}/{rid[3:6]}/{rid[6:]}"
    endpoint = dcor_shared.get_ckan_config_option(
        "dcor_object_store.endpoint_url")
    prev_url = f"{endpoint}/{bucket_name}/{object_name}"
    response = requests.get(prev_url)
    assert response.ok, "resource is public"
    assert response.status_code == 200
    # Verify SHA256sum
    path = dcor_shared.get_resource_path(res_dict["id"])
    path_prev = path.with_name(path.name + "_preview.jpg")
    dl_path = tmpdir / "prev_image.jpg"
    with dl_path.open("wb") as fd:
        fd.write(response.content)
    assert dcor_shared.sha256sum(dl_path) == dcor_shared.sha256sum(path_prev)



import copy

from flask import Blueprint
from ckan.common import config
import ckan.lib.datapreview as datapreview
from ckan.lib.jobs import _connect as ckan_redis_connect
import ckan.plugins.toolkit as toolkit
import ckan.plugins as plugins

from dcor_shared import DC_MIME_TYPES, s3
from rq.job import Job

from .cli import get_commands
from .jobs import create_preview_job, migrate_preview_to_s3_job
from .meta import render_metadata_html
from .route_funcs import dcpreview


class DCViewPlugin(plugins.SingletonPlugin):
    """DC data view and route for *_preview.png"""
    plugins.implements(plugins.IBlueprint)
    plugins.implements(plugins.IClick)
    plugins.implements(plugins.IConfigurer, inherit=True)
    plugins.implements(plugins.IResourceController, inherit=True)
    plugins.implements(plugins.IResourceView, inherit=True)

    # IBlueprint
    def get_blueprint(self):
        """"Return a Flask Blueprint object to be registered by the app."""

        # Create Blueprint for plugin
        blueprint = Blueprint(self.name, self.__module__)

        # Add plugin url rules to Blueprint object
        rules = [
            ('/dataset/<uuid:ds_id>/resource/<uuid:res_id>/preview.jpg',
             'dcpreview',
             dcpreview),
        ]
        for rule in rules:
            blueprint.add_url_rule(*rule)
        return blueprint

    # IClick
    def get_commands(self):
        return get_commands()

    # IConfigurer
    def update_config(self, config):
        plugins.toolkit.add_template_directory(config, 'templates')
        plugins.toolkit.add_resource('assets', 'dc_view')

    # IResourceController
    def after_resource_create(self, context, resource):
        """Generate preview data"""
        if resource.get('mimetype') in DC_MIME_TYPES:
            pkg_job_id = f"{resource['package_id']}_{resource['position']}_"
            depends_on = []
            extensions = [config.get("ckan.plugins")]
            # Are we waiting for symlinking (ckanext-dcor_depot)?
            # (This makes wait_for_resource really fast ;)
            if "dcor_depot" in extensions:
                # Wait for the resource to be moved to the depot.
                jid_sl = pkg_job_id + "symlink"
                depends_on.append(jid_sl)
            jid_preview = pkg_job_id + "preview"
            if not Job.exists(jid_preview, connection=ckan_redis_connect()):
                toolkit.enqueue_job(create_preview_job,
                                    [resource],
                                    title="Create resource preview image",
                                    queue="dcor-normal",
                                    rq_kwargs={
                                        "timeout": 3600,
                                        "job_id": jid_preview,
                                        "depends_on": copy.copy(depends_on)})

            # Upload the condensed dataset to S3
            if s3.is_available():
                jid_condensed_s3 = pkg_job_id + "previews3"
                toolkit.enqueue_job(
                    migrate_preview_to_s3_job,
                    [resource],
                    title="Migrate preview image to S3 object store",
                    queue="dcor-normal",
                    rq_kwargs={"timeout": 1000,
                               "job_id": jid_condensed_s3,
                               "depends_on": [jid_preview]}
                    )

    # IResourceView
    def info(self):
        return {'name': 'dc_view',
                'title': plugins.toolkit._('DC Info'),
                'icon': 'microscope',
                'iframed': False,
                'always_available': True,
                'default_title': plugins.toolkit._('DC Info'),
                }

    def can_view(self, data_dict):
        resource = data_dict['resource']
        mtype = resource.get('mimetype', '')
        same_domain = datapreview.on_same_domain(data_dict)
        if mtype in DC_MIME_TYPES and same_domain:
            return True
        else:
            return False

    def setup_template_variables(self, context, data_dict):
        preview_url = '/dataset/{}/resource/{}/preview.jpg'.format(
            data_dict['package']['id'], data_dict['resource']['id'])
        metadata_html = render_metadata_html(data_dict["resource"])
        return {
            'metadata_html': metadata_html,
            'preview_url': preview_url,
        }

    def view_template(self, context, data_dict):
        return 'dc_view.html'



import pathlib

import ckan.tests.helpers as helpers


data_path = pathlib.Path(__file__).parent / "data"


def make_dataset(create_context, owner_org, create_with_upload=None,
                 test_file_name="calibration_beads_47.rtdc",
                 activate=False, **kwargs):
    if "title" not in kwargs:
        kwargs["title"] = "test-dataset"
    if "authors" not in kwargs:
        kwargs["authors"] = "Peter Pan"
    if "license_id" not in kwargs:
        kwargs["license_id"] = "CC-BY-4.0"
    assert "state" not in kwargs, "must not be set"
    assert "owner_org" not in kwargs, "must not be set"
    # create a dataset
    ds = helpers.call_action("package_create", create_context,
                             owner_org=owner_org["name"],
                             state="draft",
                             **kwargs
                             )

    if create_with_upload is not None:
        rs = make_resource(create_with_upload, create_context, ds["id"],
                           test_file_name=test_file_name)

    if activate:
        helpers.call_action("package_patch", create_context,
                            id=ds["id"],
                            state="active")

    dataset = helpers.call_action("package_show", id=ds["id"])

    if create_with_upload is not None:
        return dataset, rs
    else:
        return dataset


def make_resource(create_with_upload, create_context, dataset_id,
                  test_file_name="calibration_beads_47.rtdc"):
    content = (data_path / test_file_name).read_bytes()
    rs = create_with_upload(
        data=content,
        filename=test_file_name,
        context=create_context,
        package_id=dataset_id,
        url="upload",
    )
    resource = helpers.call_action("resource_show", id=rs["id"])
    return resource


def synchronous_enqueue_job(job_func, args=None, kwargs=None, title=None,
                            queue=None, rq_kwargs=None):
    """
    Synchronous mock for ``ckan.plugins.toolkit.enqueue_job``.


    Due to the asynchronous nature of background jobs, code that uses them
    needs to be handled specially when writing tests.

    A common approach is to use the mock package to replace the
    ckan.plugins.toolkit.enqueue_job function with a mock that executes jobs
    synchronously instead of asynchronously

    Also, since we are running the tests as root on a ckan instance that
    is run by www-data, modifying files on disk in background jobs
    (which were started by supervisor as www-data) does not work.
    """
    if rq_kwargs is None:
        rq_kwargs = {}
    args = args or []
    kwargs = kwargs or {}
    job_func(*args, **kwargs)



import atexit
from collections import OrderedDict
import os
import shutil
import tempfile

import ckan.plugins.toolkit as toolkit
import dclab
from dclab.rtdc_dataset import linker
from dcor_shared import (
    DC_MIME_TYPES, s3, sha256sum, get_ckan_config_option, get_resource_path,
    wait_for_resource)
import numpy as np

# Create a temporary matplotlib config directory which is removed on exit
mpldir = tempfile.mkdtemp(prefix="ckan_dcor_dc_view_")
atexit.register(shutil.rmtree, mpldir)
os.environ['MPLCONFIGDIR'] = mpldir

from matplotlib.gridspec import GridSpec  # noqa: E402
import matplotlib  # noqa: E402
matplotlib.use('agg')
import matplotlib.pylab as plt  # noqa: E402


def admin_context():
    return {'ignore_auth': True, 'user': 'default'}


def create_preview_job(resource, override=False):
    """Generate a *_preview.png file for a DC resource"""
    path = get_resource_path(resource["id"])
    wait_for_resource(path)
    mtype = resource.get('mimetype', '')
    if mtype in DC_MIME_TYPES:
        # only do this for rtdc data
        jpgpath = path.with_name(path.name + "_preview.jpg")
        if not jpgpath.exists() or override:
            generate_preview(path, jpgpath)
            return True
    return False


def migrate_preview_to_s3_job(resource):
    """Migrate a preview image to the S3 object store"""
    path = get_resource_path(resource["id"])
    path_prev = path.with_name(path.name + "_preview.jpg")
    ds_dict = toolkit.get_action('package_show')(
        admin_context(),
        {'id': resource["package_id"]})
    # Perform the upload
    bucket_name = get_ckan_config_option(
        "dcor_object_store.bucket_name").format(
        organization_id=ds_dict["organization"]["id"])
    rid = resource["id"]
    sha256 = sha256sum(path_prev)
    s3.upload_file(
        bucket_name=bucket_name,
        object_name=f"preview/{rid[:3]}/{rid[3:6]}/{rid[6:]}",
        path=path_prev,
        sha256=sha256,
        private=ds_dict["private"])
    # TODO: delete the local resource after successful upload?


def generate_preview(path_rtdc, path_jpg):
    # Check whether we have a condensed version of the dataset.
    # If so, include that in the overview plot.
    paths = []
    path_condensed = path_rtdc.with_name(path_rtdc.name + "_condensed.rtdc")
    if path_condensed.exists():
        paths.append(path_condensed)
    paths.append(path_rtdc)
    # We create a linked HDF5 file which includes all paths.
    with linker.combine_h5files(paths, external="raise") as fd:
        # We do not enable basins, because it is time-consuming.
        ds = dclab.rtdc_dataset.fmt_hdf5.RTDC_HDF5(fd, enable_basins=False)
        # This is the original dataset
        fig = overview_plot(rtdc_ds=ds)
        fig.savefig(str(path_jpg), dpi=80)
        plt.close()


def overview_plot(rtdc_ds):
    """Simple overview plot adapted from the dclab examples

    Parameters
    ----------
    rtdc_ds: dclab.rtdc_dataset.core.RTDCBase
        Full RT-DC dataset to plot

    .. versionchanged:: 0.5.10

        Only the first 5000 events are plotted for performance reasons

    .. verionchanged:: 0.8.1

        Removed `rtdc_ds_cond` argument, since we are now working
        with linked datasets.

    """
    # Only plot the first 5000 events
    size = min(len(rtdc_ds), 5000)
    rtdc_ds.filter.manual[:] = False
    rtdc_ds.filter.manual[:size] = True
    rtdc_ds.apply_filter()
    ds = dclab.new_dataset(rtdc_ds)

    # Features for scatter plot
    scatter_x = "area_um"
    scatter_y = "deform"
    # Event index to display
    event_index = min(len(ds)-1, 47)

    xlabel = dclab.dfn.get_feature_label(scatter_x, rtdc_ds=ds)
    ylabel = dclab.dfn.get_feature_label(scatter_y, rtdc_ds=ds)

    plots = OrderedDict()
    plots["scatter_basic"] = scatter_x in ds and scatter_y in ds
    plots["scatter_kde"] = scatter_x in ds and scatter_y in ds
    plots["image"] = "image" in ds
    plots["mask"] = "mask" in ds
    plots["trace"] = "trace" in ds

    numplots = sum(plots.values())

    if not numplots:
        # empty plot
        fig = plt.figure(figsize=(4, 4))
        return fig

    height_ratios = []
    for key in plots:
        if plots[key]:
            if key in ["image", "mask"]:
                height_ratios.append(1)
            else:
                height_ratios.append(2)

    fig = plt.figure(figsize=(4, np.sum(height_ratios)*1.5))

    gs = GridSpec(numplots, 1, height_ratios=height_ratios)
    ii = 0

    if scatter_x in ds and scatter_y in ds:
        ax1 = fig.add_subplot(gs[ii])
        ax1.set_title("Basic scatter plot")
        ii += 1
        x_start = np.percentile(ds[scatter_x], 1)
        x_end = np.percentile(ds[scatter_x], 99)
        y_start = np.percentile(ds[scatter_y], 1)
        y_end = np.percentile(ds[scatter_y], 99)

        ax1.plot(ds[scatter_x], ds[scatter_y],
                 "o", color="k", alpha=.2, ms=1)
        ax1.set_xlabel(xlabel)
        ax1.set_ylabel(ylabel)
        ax1.set_xlim(x_start, x_end)
        ax1.set_ylim(y_start, y_end)

        ax2 = fig.add_subplot(gs[ii])
        ax2.set_title("KDE scatter plot")
        ii += 1
        sc = ax2.scatter(ds[scatter_x], ds[scatter_y],
                         c=ds.get_kde_scatter(xax=scatter_x,
                                              yax=scatter_y,
                                              kde_type="histogram"),
                         s=3)
        plt.colorbar(sc, label="kernel density [a.u]", ax=ax2)
        ax2.set_xlabel(xlabel)
        ax2.set_ylabel(ylabel)
        ax2.set_xlim(x_start, x_end)
        ax2.set_ylim(y_start, y_end)

    if "image" in ds:
        ax3 = fig.add_subplot(gs[ii])
        ax3.set_title("Event image with contour")
        ii += 1
        ax3.imshow(ds["image"][event_index], cmap="gray")
        ax3.set_xlabel("Detector X [px]")
        ax3.set_ylabel("Detector Y [px]")

        if "contour" in ds:
            ax3.plot(ds["contour"][event_index][:, 0],
                     ds["contour"][event_index][:, 1],
                     c="r")

    if "mask" in ds:
        ax4 = fig.add_subplot(gs[ii])
        ax4.set_title("Event mask")
        ii += 1
        pxsize = ds.config["imaging"]["pixel size"]
        ax4.imshow(ds["mask"][event_index],
                   extent=(0, ds["mask"][0].shape[1] * pxsize,
                           0, ds["mask"][0].shape[0] * pxsize),
                   cmap="gray")
        ax4.set_xlabel(u"Detector X [µm]")
        ax4.set_ylabel(u"Detector Y [µm]")

    if "trace" in ds:
        ax5 = fig.add_subplot(gs[ii])
        ax5.set_title("Fluorescence traces")
        ii += 1
        flsamples = ds.config["fluorescence"]["samples per event"]
        flrate = ds.config["fluorescence"]["sample rate"]
        fltime = np.arange(flsamples) / flrate * 1e6
        if "fl1_raw" in ds["trace"]:
            ax5.plot(fltime, ds["trace"]["fl1_raw"][event_index],
                     c="#15BF00", label="fl1_raw")
        if "fl2_raw" in ds["trace"]:
            ax5.plot(fltime, ds["trace"]["fl2_raw"][event_index],
                     c="#BF8A00", label="fl2_raw")
        if "fl3_raw" in ds["trace"]:
            ax5.plot(fltime, ds["trace"]["fl3_raw"][event_index],
                     c="#BF0C00", label="fl3_raw")
        ax5.legend()
        ax5.set_xlabel(u"Event time [µs]")
        ax5.set_ylabel("Fluorescence [a.u.]")
        ax5.set_xlim(0, fltime[-1])

    plt.tight_layout()

    return fig



import pathlib

import flask
from ckan.common import c
import ckan.lib.uploader as uploader
from ckan import logic
import ckan.model as model
import ckan.plugins.toolkit as toolkit

import botocore.exceptions
from dcor_shared import get_ckan_config_option, s3


def dcpreview(ds_id, res_id):
    """Serve a preview image on disk

    Parameters
    ----------
    ds_id: str
        dataset ID
    res_id: str
        resource ID for which to return the preview image
    """
    # Code borrowed from ckan/controllers/package.py:resource_download
    context = {'model': model, 'session': model.Session,
               'user': c.user, 'auth_user_obj': c.userobj}
    did = str(ds_id)
    rid = str(res_id)
    try:
        res_dict = toolkit.get_action('resource_show')(context, {'id': rid})
        ds_dict = toolkit.get_action('package_show')(context, {'id': did})
    except (logic.NotFound, logic.NotAuthorized):
        # Treat not found and not authorized equally, to not leak information
        # to unprivileged users.
        return toolkit.abort(404, toolkit._('Resource not found'))

    res_stem, _ = res_dict["name"].rsplit(".", 1)
    prev_name = f"{res_stem}_preview.jpg"

    if s3 is not None and res_dict.get('s3_available'):
        # check if the corresponding S3 object exists
        bucket_name = get_ckan_config_option(
            "dcor_object_store.bucket_name").format(
            organization_id=ds_dict["organization"]["id"])
        object_name = f"preview/{rid[:3]}/{rid[3:6]}/{rid[6:]}"
        s3_client, _, _ = s3.get_s3()
        try:
            s3_client.head_object(Bucket=bucket_name,
                                  Key=object_name)
        except botocore.exceptions.ClientError:
            pass
        else:
            # We have an S3 object that we can redirect to. We are making use
            # of presigned URLs to be able to specify a filename for download
            # (otherwise, users that download via the web interface will
            # just get a hash as a file name without any suffix or human-
            # readable identifier).
            if ds_dict["private"]:
                expiration = 3600
            else:
                expiration = 86400
            ps_url = s3.create_presigned_url(
                bucket_name=bucket_name,
                object_name=object_name,
                filename=prev_name,
                expiration=expiration)
            return toolkit.redirect_to(ps_url)

    if res_dict.get('url_type') == 'upload':
        upload = uploader.get_resource_uploader(res_dict)
        filepath = pathlib.Path(upload.get_path(res_dict['id']))
        jpg_file = filepath.with_name(filepath.name + "_preview.jpg")
        if not jpg_file.exists():
            return toolkit.abort(404, toolkit._('Preview not found'))
        return flask.send_from_directory(jpg_file.parent, jpg_file.name,
                                         attachment_filename=prev_name)
    return toolkit.abort(404, toolkit._('No preview available'))



from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
from codecs import open  # To use a consistent encoding
import sys
import warnings

author = "Paul Müller"
authors = [author]
description = 'Preview DC data in DCOR '
name = 'ckanext-dc_view'
year = "2020"

sys.path.insert(0, realpath(dirname(__file__))+"/" + "/".join(name.split("-")))
from _version import version  # noqa: E402

try:
    # Make sure this fails for old CKAN versions
    import ckan
    ckan_version = [int(v) for v in ckan.__version__.split(".")]
    if ckan_version < [2, 10, 1]:
        raise ValueError(
            f"Your CKAN version {ckan_version} is not supported! If you "
            f"are still on CKAN 2.9.5, then the following package versions "
            f"are supported:"
            f"\n ckanext-dc_view<=0.6.10"
            )
except ImportError:
    warnings.warn("CKAN not installed, supported version check skipped.")

setup(
    name=name,
    version=version,
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    url='https://github.com/DCOR-dev/ckanext-dc_view',
    author=author,
    author_email='dev@craban.de',
    license='AGPLv3+',
    keywords=["CKAN", "DCOR", "RT-DC"],
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    package_dir={name: name},
    namespace_packages=['ckanext'],
    install_requires=[
        # the "ckan" dependency is implied
        "dclab>=0.57.0",
        "dcor_shared>=0.5.3",
        "matplotlib",
        "numpy>=1.19",
        "pillow",
    ],
    include_package_data=True,
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points='''
        [ckan.plugins]
        dc_view=ckanext.dc_view.plugin:DCViewPlugin
    ''',
    # If you are changing from the default layout of your extension, you may
    # have to change the message extractors, you can read more about babel
    # message extraction at
    # http://babel.pocoo.org/docs/messages/#extraction-method-mapping-and-configuration
    message_extractors={
        'ckanext': [
            ('**.py', 'python', None),
            ('**.js', 'javascript', None),
            ('**/templates/**.html', 'ckan', None),
        ],
    },
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'License :: OSI Approved :: GNU Affero General Public License v3 or ' \
        + 'later (AGPLv3+)',
        'Programming Language :: Python :: 3',
    ],
)



import numbers

import dclab


def render_metadata_html(res_dict):
    # build dictionary from metadata
    meta = {}
    for dckey in res_dict:
        if dckey.startswith("dc:"):
            _, sec, key = dckey.split(":")
            if sec not in meta:
                meta[sec] = {}
            meta[sec][key] = res_dict[dckey]

    html = []
    for sec in ["experiment", "setup", "imaging", "fluorescence"]:
        if sec in meta:
            html += meta_html_table(meta, sec)
            html.append("<br>")

    return "\n".join(html)


def meta_html_table(meta, sec):
    html = [
        '<table class="table table-striped table-bordered table-condensed">',
        "<caption>{}</caption>".format(sec.capitalize()),
    ]

    kn = [(dclab.dfn.config_descr[sec][key], key) for key in meta[sec].keys()]
    for name, key in sorted(kn):
        value = meta[sec][key]
        if isinstance(value, numbers.Number):
            value = "{:.4g}".format(value)

        # Special cases
        if sec == "experiment":
            if key in ["date", "time"]:
                name, _ = name.split("(")
            elif key == "sample":
                name = "Sample name"
        elif sec == "setup":
            if key == "chip region":
                name = name.split(" (")[0]
            elif key == "medium" and value == "CellCarrierB":
                value = "CellCarrier B"
            elif key == "module composition":
                name = "Modules used"
                value = ", ".join(value.split(","))
            elif key == "software version":
                name = "Software"

        # Units
        if name.endswith("]"):
            name, units = name.rsplit(" [", 1)
            units = units.strip("] ")
            value += " " + units

        html += [
            '<tr>',
            u'<th class="dataset-labels">{}</th>'.format(name),
            u'<td class="dataset-details">{}</td>'.format(value),
            '</tr>',
        ]
    html.append("</table>")
    return html



import mock

import ckan.common
import ckan.model
import ckan.tests.factories as factories
import ckan.tests.helpers as helpers
import ckanext.dcor_schemas.plugin
import dcor_shared

import pytest

from .helper_methods import make_dataset, synchronous_enqueue_job


@pytest.mark.ckan_config('ckan.plugins',
                         'dcor_depot dcor_schemas dc_serve dc_view')
@pytest.mark.usefixtures('clean_db', 'with_request_context')
@mock.patch('ckan.plugins.toolkit.enqueue_job',
            side_effect=synchronous_enqueue_job)
def test_route_redircet_preview_to_s3_private(
        enqueue_job_mock, app, tmpdir, create_with_upload, monkeypatch,
        ckan_config):
    monkeypatch.setitem(ckan_config, 'ckan.storage_path', str(tmpdir))
    monkeypatch.setattr(ckan.lib.uploader,
                        'get_storage_path',
                        lambda: str(tmpdir))
    monkeypatch.setattr(
        ckanext.dcor_schemas.plugin,
        'DISABLE_AFTER_DATASET_CREATE_FOR_CONCURRENT_JOB_TESTS',
        True)

    user = factories.User()
    user_obj = ckan.model.User.by_name(user["name"])
    monkeypatch.setattr(ckan.common,
                        'current_user',
                        user_obj)
    owner_org = factories.Organization(users=[{
        'name': user['id'],
        'capacity': 'admin'
    }])
    # Note: `call_action` bypasses authorization!
    create_context = {'ignore_auth': False,
                      'user': user['name'],
                      'api_version': 3}
    # create a dataset
    ds_dict, res_dict = make_dataset(create_context, owner_org,
                                     create_with_upload=create_with_upload,
                                     activate=True,
                                     private=True
                                     )
    rid = res_dict["id"]
    assert "s3_available" in res_dict
    assert "s3_url" in res_dict

    # Remove the local resource to make sure CKAN serves the S3 URL
    path = dcor_shared.get_resource_path(rid)
    path_prev = path.with_name(path.name + "_preview.jpg")
    assert path_prev.exists()
    path_prev.unlink()

    did = ds_dict["id"]
    # We should not be authorized to access the resource without API token
    resp0 = app.get(
        f"/dataset/{did}/resource/{rid}/preview.jpg",
        status=404
        )
    assert len(resp0.history) == 0

    # Try again with token
    data = helpers.call_action(
        u"api_token_create",
        context={u"model": ckan.model, u"user": user[u"name"]},
        user=user[u"name"],
        name=u"token-name",
    )

    resp = app.get(
        f"/dataset/{did}/resource/{rid}/preview.jpg",
        headers={u"authorization": data["token"]},
        )

    endpoint = dcor_shared.get_ckan_config_option(
        "dcor_object_store.endpoint_url")
    bucket_name = dcor_shared.get_ckan_config_option(
        "dcor_object_store.bucket_name").format(
        organization_id=ds_dict["organization"]["id"])
    redirect = resp.history[0]
    assert redirect.status_code == 302
    redirect_stem = (f"{endpoint}/{bucket_name}/preview/"
                     f"{rid[:3]}/{rid[3:6]}/{rid[6:]}")
    # Since we have a presigned URL, it is longer than the normal S3 URL.
    assert redirect.location.startswith(redirect_stem)
    assert len(redirect.location) > len(redirect_stem)


@pytest.mark.ckan_config('ckan.plugins',
                         'dcor_depot dcor_schemas dc_serve dc_view')
@pytest.mark.usefixtures('clean_db', 'with_request_context')
@mock.patch('ckan.plugins.toolkit.enqueue_job',
            side_effect=synchronous_enqueue_job)
def test_route_preview_to_s3_public(
        enqueue_job_mock, app, tmpdir, create_with_upload, monkeypatch,
        ckan_config):
    monkeypatch.setitem(ckan_config, 'ckan.storage_path', str(tmpdir))
    monkeypatch.setattr(ckan.lib.uploader,
                        'get_storage_path',
                        lambda: str(tmpdir))
    monkeypatch.setattr(
        ckanext.dcor_schemas.plugin,
        'DISABLE_AFTER_DATASET_CREATE_FOR_CONCURRENT_JOB_TESTS',
        True)

    user = factories.User()
    user_obj = ckan.model.User.by_name(user["name"])
    monkeypatch.setattr(ckan.common,
                        'current_user',
                        user_obj)
    owner_org = factories.Organization(users=[{
        'name': user['id'],
        'capacity': 'admin'
    }])
    # Note: `call_action` bypasses authorization!
    create_context = {'ignore_auth': False,
                      'user': user['name'],
                      'api_version': 3}
    # create a dataset
    ds_dict, res_dict = make_dataset(create_context, owner_org,
                                     create_with_upload=create_with_upload,
                                     activate=True)
    rid = res_dict["id"]
    assert "s3_available" in res_dict
    assert "s3_url" in res_dict

    # Remove the local resource to make sure CKAN serves the S3 URL
    path = dcor_shared.get_resource_path(rid)
    path_prev = path.with_name(path.name + "_preview.jpg")
    assert path_prev.exists()
    path_prev.unlink()

    did = ds_dict["id"]
    resp = app.get(
        f"/dataset/{did}/resource/{rid}/preview.jpg",
        )

    endpoint = dcor_shared.get_ckan_config_option(
        "dcor_object_store.endpoint_url")
    bucket_name = dcor_shared.get_ckan_config_option(
        "dcor_object_store.bucket_name").format(
        organization_id=ds_dict["organization"]["id"])
    redirect = resp.history[0]
    assert redirect.status_code == 302
    assert redirect.location.startswith(f"{endpoint}/{bucket_name}/preview/"
                                        f"{rid[:3]}/{rid[3:6]}/{rid[6:]}")
